import time
import logging
from typing import Optional, Union, Iterator

from ..strategy.strategy import BaseStrategy
from .marketstream import MarketStream
from .sportsdatastream import SportsDataStream
from .datastream import DataStream
from .historicalstream import HistoricalStream
from .orderstream import OrderStream
from .simulatedorderstream import SimulatedOrderStream
from ..clients import ExchangeType, BaseClient
from ..utils import get_file_md
from betfairlightweight.resources.streamingresources import MarketDefinition

logger = logging.getLogger(__name__)


class Streams:
    def __init__(self, flumine):
        self.flumine = flumine
        self._streams = []
        self._stream_id = 0

    def __call__(self, strategy: BaseStrategy) -> None:
        if self.flumine.SIMULATED:
            markets = strategy.market_filter.get("markets")
            market_types = strategy.market_filter.get("market_types")
            country_codes = strategy.market_filter.get("country_codes")
            event_processing = strategy.market_filter.get("event_processing", False)
            events = strategy.market_filter.get("events")
            listener_kwargs = strategy.market_filter.get("listener_kwargs", {})
            if markets and events:
                logger.warning(
                    "Markets and events found for strategy %s skipping as flumine can only handle one type",
                    strategy,
                )
            elif markets is None and events is None:
                logger.warning("No markets or events found for strategy %s", strategy)
            elif markets:
                # order markets by name as an attempt to process in chronological order
                markets.sort()
                for market in markets:
                    market_definition = get_file_md(market)
                    market_type = getattr(market_definition, "market_type", None)
                    country_code = getattr(market_definition, "country_code", None)
                    if market_types and market_type and market_type not in market_types:
                        logger.warning(
                            "Skipping market %s (%s) for strategy %s due to marketType filter",
                            market,
                            market_type,
                            strategy,
                        )
                    elif (
                        country_codes
                        and country_code
                        and country_code not in country_codes
                    ):
                        logger.warning(
                            "Skipping market %s (%s) for strategy %s due to countryCode filter"
                            % (market, country_code, strategy)
                        )
                    else:
                        stream = self.add_historical_stream(
                            strategy,
                            market,
                            market_definition,
                            event_processing,
                            **listener_kwargs,
                        )
                        strategy.streams.append(stream)
                        strategy.historic_stream_ids.add(stream.stream_id)
            elif events:
                raise NotImplementedError()
        else:
            self.add_stream(strategy)

    def add_client(self, client: BaseClient) -> None:
        if client.order_stream:
            if client.paper_trade:
                self.add_simulated_order_stream(client)
            elif client.EXCHANGE == ExchangeType.BETFAIR:
                self.add_order_stream(client)

    """ market data """

    def add_stream(self, strategy: BaseStrategy) -> None:
        # markets
        if isinstance(strategy.market_filter, dict) or strategy.market_filter is None:
            market_filters = [strategy.market_filter]
        else:
            market_filters = strategy.market_filter
        for market_filter in market_filters:
            for stream in self:  # check if market stream already exists
                if (
                    isinstance(stream, strategy.stream_class)
                    and stream.market_filter == market_filter
                    and stream.market_data_filter == strategy.market_data_filter
                    and stream.streaming_timeout == strategy.streaming_timeout
                    and stream.conflate_ms == strategy.conflate_ms
                ):
                    logger.info(
                        "Using %s (%s) for strategy %s",
                        strategy.stream_class,
                        stream.stream_id,
                        strategy,
                    )
                    strategy.streams.append(stream)
                    break
            else:  # nope? lets create a new one
                stream_id = self._increment_stream_id()
                logger.info(
                    "Creating new %s (%s) for strategy %s",
                    strategy.stream_class,
                    stream_id,
                    strategy,
                )
                stream = strategy.stream_class(
                    flumine=self.flumine,
                    stream_id=stream_id,
                    market_filter=market_filter,
                    market_data_filter=strategy.market_data_filter,
                    streaming_timeout=strategy.streaming_timeout,
                    conflate_ms=strategy.conflate_ms,
                )
                self._streams.append(stream)
                strategy.streams.append(stream)
        # sports data
        for subscription in strategy.sports_data_filter:
            for stream in self:  # check if sports data stream already exists
                if (
                    isinstance(stream, SportsDataStream)
                    and stream.sports_data_filter == subscription
                    and stream.streaming_timeout == strategy.streaming_timeout
                ):
                    logger.info(
                        "Using %s (%s) for strategy %s",
                        strategy.stream_class,
                        stream.stream_id,
                        strategy,
                    )
                    strategy.streams.append(stream)
                    break
            else:  # nope? lets create a new one
                stream_id = self._increment_stream_id()
                logger.info(
                    "Creating new %s (%s) for strategy %s",
                    strategy.stream_class,
                    stream_id,
                    strategy,
                )
                stream = SportsDataStream(
                    flumine=self.flumine,
                    stream_id=stream_id,
                    sports_data_filter=subscription,
                    streaming_timeout=strategy.streaming_timeout,
                )
                self._streams.append(stream)
                strategy.streams.append(stream)

    def add_historical_stream(
        self,
        strategy: BaseStrategy,
        market: str,
        market_definition: Optional[MarketDefinition],
        event_processing: bool,
        **listener_kwargs,
    ) -> HistoricalStream:
        for stream in self:
            if (
                stream.market_filter == market
                and stream.event_processing == event_processing
                and stream.listener_kwargs == listener_kwargs
            ):
                return stream
        else:
            stream_id = self._increment_stream_id()
            event_id = getattr(market_definition, "event_id", None)
            if event_processing and event_id is None:
                logger.warning("EventId not found for market %s" % market)
            logger.info(
                "Creating new %s (%s) for strategy %s",
                HistoricalStream.__name__,
                stream_id,
                strategy,
                extra={
                    "strategy": strategy,
                    "stream_id": stream_id,
                    "market_filter": market,
                    "event_id": event_id,
                    "event_processing": event_processing,
                },
            )
            stream = HistoricalStream(
                flumine=self.flumine,
                stream_id=stream_id,
                market_filter=market,
                market_data_filter=strategy.market_data_filter,
                streaming_timeout=strategy.streaming_timeout,
                conflate_ms=strategy.conflate_ms,
                output_queue=False,
                event_processing=event_processing,
                event_id=event_id,
                **listener_kwargs,
            )
            self._streams.append(stream)
            return stream

    """ order data """

    def add_order_stream(
        self,
        client: BaseClient,
        conflate_ms: int = None,
        streaming_timeout: float = 0.25,
    ) -> OrderStream:
        stream_id = self._increment_stream_id()
        stream = OrderStream(
            flumine=self.flumine,
            stream_id=stream_id,
            conflate_ms=conflate_ms,
            streaming_timeout=streaming_timeout,
            client=client,
        )
        self._streams.append(stream)
        return stream

    def add_simulated_order_stream(
        self,
        client: BaseClient,
        conflate_ms: int = None,
        streaming_timeout: float = 0.25,
    ) -> SimulatedOrderStream:
        logger.warning("Client %s now paper trading", client.betting_client.username)
        stream_id = self._increment_stream_id()
        stream = SimulatedOrderStream(
            flumine=self.flumine,
            stream_id=stream_id,
            conflate_ms=conflate_ms,
            streaming_timeout=streaming_timeout,
            client=client,
            custom=True,
        )
        self._streams.append(stream)
        return stream

    """ custom stream """

    def add_custom_stream(self, stream):
        stream.stream_id = self._increment_stream_id()
        self._streams.append(stream)
        return stream

    def start(self) -> None:
        if not self.flumine.SIMULATED:
            logger.info("Starting streams..")
            for stream in self:
                stream.start()
                # wait for successful start
                while not stream.custom and not stream.stream_running:
                    time.sleep(0.25)

    def stop(self) -> None:
        for stream in self:
            stream.stop()

    def _increment_stream_id(self) -> int:
        self._stream_id += int(1e3)
        return self._stream_id

    def __iter__(self) -> Iterator[Union[MarketStream, DataStream, HistoricalStream]]:
        return iter(self._streams)

    def __len__(self) -> int:
        return len(self._streams)



import logging
from typing import Optional
from tenacity import retry
from betfairlightweight import StreamListener, filters, BetfairError
from betfairlightweight.streaming.stream import BaseStream as BFBaseStream

from .basestream import BaseStream
from ..events.events import RawDataEvent
from .. import config

logger = logging.getLogger(__name__)

RETRY_WAIT = BaseStream.RETRY_WAIT

"""
Custom listener that doesn't do any processing,
helps reduce CPU.
"""


class FlumineListener(StreamListener):
    def _add_stream(self, unique_id: int, operation: str) -> BFBaseStream:
        if operation == "marketSubscription":
            return FlumineMarketStream(self, unique_id)
        elif operation == "orderSubscription":
            return FlumineOrderStream(self, unique_id)
        elif operation == "raceSubscription":
            return FlumineRaceStream(self, unique_id)
        elif operation == "cricketSubscription":
            return FlumineCricketStream(self, unique_id)


class FlumineStream(BFBaseStream):
    def on_process(self, caches: list, publish_time: Optional[int] = None) -> None:
        output = RawDataEvent(caches)
        self.output_queue.put(output)

    def __str__(self):
        return "FlumineStream"

    def __repr__(self):
        return "<FlumineStream [%s]>" % len(self._caches)


class FlumineMarketStream(FlumineStream):
    _lookup = "mc"

    def _process(self, data: list, publish_time: int) -> bool:
        for market_book in data:
            market_id = market_book.get("id")
            if (
                "marketDefinition" in market_book
                and market_book["marketDefinition"]["status"] == "CLOSED"
            ):
                if market_id in self._caches:
                    # removes closed market from cache
                    del self._caches[market_id]
                    logger.info(
                        "[MarketStream: %s] %s removed, %s markets in cache",
                        self.unique_id,
                        market_id,
                        len(self._caches),
                    )
            elif market_id not in self._caches:
                # adds empty object to cache to track live market count
                self._caches[market_id] = object()
                logger.info(
                    "[MarketStream: %s] %s added, %s markets in cache",
                    self.unique_id,
                    market_id,
                    len(self._caches),
                )
            self._updates_processed += 1

        self.on_process([self.unique_id, self._clk, publish_time, data])
        return False


class FlumineOrderStream(FlumineStream):
    _lookup = "oc"

    def _process(self, data: list, publish_time: int) -> bool:
        for update in data:
            market_id = update["id"]
            if market_id not in self._caches:
                # adds empty object to cache to track live market count
                self._caches[market_id] = object()
                logger.info(
                    "[OrderStream: %s] %s added, %s markets in cache",
                    self.unique_id,
                    market_id,
                    len(self._caches),
                )
            self._updates_processed += 1

        self.on_process([self.unique_id, self._clk, publish_time, data])
        return False


class FlumineRaceStream(FlumineStream):
    _lookup = "rc"

    def _process(self, data: list, publish_time: int) -> bool:
        for update in data:
            market_id = update["mid"]
            if market_id not in self._caches:
                # adds empty object to cache to track live market count
                self._caches[market_id] = object()
                logger.info(
                    "[RaceStream: %s] %s added, %s markets in cache",
                    self.unique_id,
                    market_id,
                    len(self._caches),
                )
            self._updates_processed += 1

        self.on_process([self.unique_id, self._clk, publish_time, data])
        return False


class FlumineCricketStream(FlumineStream):
    _lookup = "cc"

    def _process(self, data: list, publish_time: int) -> bool:
        for update in data:
            market_id = update["marketId"]
            if market_id not in self._caches:
                # adds empty object to cache to track live market count
                self._caches[market_id] = object()
                logger.info(
                    "[CricketStream: %s] %s added, %s markets in cache",
                    self.unique_id,
                    market_id,
                    len(self._caches),
                )
            self._updates_processed += 1

        self.on_process([self.unique_id, self._clk, publish_time, data])
        return False


class DataStream(BaseStream):
    LISTENER = FlumineListener

    def __init__(self, *args, **kwargs):
        BaseStream.__init__(self, *args, **kwargs)
        self._listener = self.LISTENER(output_queue=self.flumine.handler_queue)

    @retry(wait=RETRY_WAIT)
    def run(self) -> None:
        logger.info(
            "Starting DataStream %s",
            self.stream_id,
            extra={
                "stream_id": self.stream_id,
                "market_filter": self.market_filter,
                "market_data_filter": self.market_data_filter,
                "conflate_ms": self.conflate_ms,
            },
        )
        self._stream = self.betting_client.streaming.create_stream(
            unique_id=self.stream_id, listener=self._listener
        )
        try:
            self.stream_id = self._stream.subscribe_to_markets(
                market_filter=self.market_filter,
                market_data_filter=self.market_data_filter,
                conflate_ms=self.conflate_ms,
                initial_clk=self._listener.initial_clk,  # supplying these two values allows a reconnect
                clk=self._listener.clk,
            )
            self._stream.start()
        except BetfairError:
            logger.error("DataStream %s run error", self.stream_id, exc_info=True)
            raise
        except Exception:
            logger.critical("DataStream %s run error", self.stream_id, exc_info=True)
            raise
        logger.info("Stopped DataStream %s", self.stream_id)


class OrderDataStream(DataStream):
    @retry(wait=RETRY_WAIT)
    def run(self) -> None:
        logger.info(
            "Starting OrderDataStream %s",
            self.stream_id,
            extra={
                "stream_id": self.stream_id,
                "customer_strategy_refs": config.customer_strategy_ref,
                "conflate_ms": self.conflate_ms,
                "streaming_timeout": self.streaming_timeout,
            },
        )
        self._stream = self.betting_client.streaming.create_stream(
            unique_id=self.stream_id, listener=self._listener
        )
        try:
            self.stream_id = self._stream.subscribe_to_orders(
                order_filter=filters.streaming_order_filter(
                    customer_strategy_refs=(
                        [config.customer_strategy_ref]
                        if config.customer_strategy_ref
                        else None
                    ),
                    partition_matched_by_strategy_ref=True,
                    include_overall_position=False,
                ),
                conflate_ms=self.conflate_ms,
            )
            self._stream.start()
        except BetfairError:
            logger.error("OrderDataStream %s run error", self.stream_id, exc_info=True)
            raise
        except Exception:
            logger.critical(
                "OrderDataStream %s run error", self.stream_id, exc_info=True
            )
            raise
        logger.info("Stopped OrderDataStream %s", self.stream_id)


class RaceDataStream(DataStream):
    @retry(wait=RETRY_WAIT)
    def run(self) -> None:
        logger.info(
            "Starting RaceDataStream %s",
            self.stream_id,
            extra={
                "stream_id": self.stream_id,
            },
        )
        self._stream = self.betting_client.streaming.create_stream(
            unique_id=self.stream_id, listener=self._listener, host="sports_data"
        )
        try:
            self.stream_id = self._stream.subscribe_to_races()
            self._stream.start()
        except BetfairError:
            logger.error("RaceDataStream %s run error", self.stream_id, exc_info=True)
            raise
        except Exception:
            logger.critical(
                "RaceDataStream %s run error", self.stream_id, exc_info=True
            )
            raise
        logger.info("Stopped RaceDataStream %s", self.stream_id)


class CricketDataStream(DataStream):
    @retry(wait=RETRY_WAIT)
    def run(self) -> None:
        logger.info(
            "Starting CricketDataStream %s",
            self.stream_id,
            extra={
                "stream_id": self.stream_id,
            },
        )
        self._stream = self.betting_client.streaming.create_stream(
            unique_id=self.stream_id, listener=self._listener, host="sports_data"
        )
        try:
            self.stream_id = self._stream.subscribe_to_cricket_matches()
            self._stream.start()
        except BetfairError:
            logger.error(
                "CricketDataStream %s run error", self.stream_id, exc_info=True
            )
            raise
        except Exception:
            logger.critical(
                "CricketDataStream %s run error", self.stream_id, exc_info=True
            )
            raise
        logger.info("Stopped CricketDataStream %s", self.stream_id)



import datetime
import logging
from typing import Optional
from collections import defaultdict
from pathlib import Path
from betfairlightweight.resources.bettingresources import MarketBook, MarketCatalogue

from .. import config
from .blotter import Blotter
from ..execution.transaction import Transaction

logger = logging.getLogger(__name__)


class Market:
    """
    Market data structure to hold latest marketBook,
    marketCatalogue and various properties. Also
    allows order placement through the Transaction
    class.
    """

    def __init__(
        self,
        flumine,
        market_id: str,
        market_book: MarketBook,
        market_catalogue: MarketCatalogue = None,
    ):
        self.flumine = flumine
        self.market_id = market_id
        self.closed = False
        self.date_time_created = datetime.datetime.utcnow()
        self.date_time_closed = None
        self.market_book = market_book
        self.market_catalogue = market_catalogue
        self.update_market_catalogue = True
        self.orders_cleared = []
        self.market_cleared = []
        self.context = {"simulated": {}}  # data store (raceCard / scores etc)
        self.blotter = Blotter(market_id)
        self._transaction_id = 0

    def __call__(self, market_book: MarketBook):
        if self.market_book and market_book.version != self.market_book.version:
            self.update_market_catalogue = True
        self.market_book = market_book

    def open_market(self) -> None:
        self.closed = False
        self.orders_cleared = []
        self.market_cleared = []
        logger.info(
            "Market %s opened",
            self.market_id,
            extra=self.info,
        )

    def close_market(self) -> None:
        self.closed = True
        self.date_time_closed = datetime.datetime.utcnow()
        logger.info(
            "Market %s closed",
            self.market_id,
            extra=self.info,
        )

    def transaction(self, async_place_orders: bool = None, client=None) -> Transaction:
        if async_place_orders is None:
            async_place_orders = config.async_place_orders
        if client is None:
            client = self.flumine.clients.get_default()
        self._transaction_id += 1
        return Transaction(
            self,
            id_=self._transaction_id,
            async_place_orders=async_place_orders,
            client=client,
        )

    # order
    def place_order(
        self,
        order,
        market_version: int = None,
        execute: bool = True,
        force: bool = False,
        client=None,
    ) -> bool:
        with self.transaction(client=client) as t:
            return t.place_order(order, market_version, execute, force)

    def cancel_order(
        self, order, size_reduction: float = None, force: bool = False
    ) -> bool:
        with self.transaction(client=order.client) as t:
            return t.cancel_order(order, size_reduction, force)

    def update_order(
        self, order, new_persistence_type: str, force: bool = False
    ) -> bool:
        with self.transaction(client=order.client) as t:
            return t.update_order(order, new_persistence_type, force)

    def replace_order(
        self, order, new_price: float, market_version: int = None, force: bool = False
    ) -> bool:
        with self.transaction(client=order.client) as t:
            return t.replace_order(order, new_price, market_version, force)

    @property
    def event(self) -> dict:
        event = defaultdict(list)
        market_start_datetime = self.market_start_datetime
        event_type_id = self.event_type_id
        for market in self.flumine.markets.events[self.event_id]:
            if (
                market_start_datetime == market.market_start_datetime
                or event_type_id in ["1", "3"]  # soccer, golf
            ):
                event[market.market_type].append(market)
        return event

    @property
    def event_type_id(self) -> str:
        if self.market_catalogue:
            return self.market_catalogue.event_type.id
        elif self.market_book:
            return self.market_book.market_definition.event_type_id

    @property
    def event_id(self) -> str:
        if self.market_catalogue:
            return self.market_catalogue.event.id
        elif self.market_book:
            return self.market_book.market_definition.event_id

    @property
    def market_type(self) -> str:
        if self.market_catalogue and self.market_catalogue.description:
            return self.market_catalogue.description.market_type
        elif self.market_book:
            return self.market_book.market_definition.market_type

    @property
    def seconds_to_start(self) -> float:
        return (self.market_start_datetime - datetime.datetime.utcnow()).total_seconds()

    @property
    def elapsed_seconds_closed(self) -> Optional[float]:
        if self.closed and self.date_time_closed:
            return (datetime.datetime.utcnow() - self.date_time_closed).total_seconds()

    @property
    def market_start_datetime(self):
        if self.market_book:
            return self.market_book.market_definition.market_time
        elif self.market_catalogue:
            return self.market_catalogue.market_start_time
        else:
            return datetime.datetime.utcfromtimestamp(0)

    @property
    def market_start_hour_minute(self) -> Optional[str]:
        msd = self.market_start_datetime
        if msd:
            return "{0:02d}{1:02d}".format(msd.hour, msd.minute)

    @property
    def event_name(self) -> Optional[str]:
        if self.market_catalogue:
            return self.market_catalogue.event.name
        elif self.market_book:
            return self.market_book.market_definition.event_name

    @property
    def country_code(self) -> Optional[str]:
        if self.market_catalogue:
            return self.market_catalogue.event.country_code
        elif self.market_book:
            return self.market_book.market_definition.country_code

    @property
    def venue(self) -> Optional[str]:
        if self.market_catalogue:
            return self.market_catalogue.event.venue
        elif self.market_book:
            return self.market_book.market_definition.venue

    @property
    def race_type(self) -> Optional[str]:
        if self.market_catalogue and self.market_catalogue.description:
            return self.market_catalogue.description.race_type
        elif self.market_book:
            return self.market_book.market_definition.race_type

    @property
    def status(self) -> Optional[str]:
        if self.market_book:
            return self.market_book.status

    def cleared(self, client) -> dict:
        orders = self.blotter.client_orders(client, matched_only=True)
        profit = round(sum([order.profit for order in orders]), 2)
        return {
            "marketId": self.market_id,
            "eventId": self.event_id,
            "eventTypeId": self.event_type_id,
            "customerStrategyRef": config.customer_strategy_ref,
            "lastMatchedDate": None,
            "placedDate": None,
            "settledDate": None,
            "betCount": len(orders),
            "betOutcome": "WON" if profit >= 0 else "LOST",
            "commission": round(max(profit * client.commission_base, 0), 2),
            "profit": profit,
        }

    @property
    def info(self) -> dict:
        return {
            "market_id": self.market_id,
            "event_id": self.event_id,
            "event_type_id": self.event_type_id,
            "event_name": self.event_name,
            "market_type": self.market_type,
            "market_start_datetime": str(self.market_start_datetime),
            "country_code": self.country_code,
            "venue": self.venue,
            "race_type": self.race_type,
            "orders_cleared": self.orders_cleared,
            "market_cleared": self.market_cleared,
            "closed": self.closed,
        }



import logging
from typing import Iterable, Optional, List
from collections import defaultdict

from ..order.ordertype import OrderTypes
from ..utils import (
    calculate_unmatched_exposure,
    calculate_matched_exposure,
    STRATEGY_NAME_HASH_LENGTH,
)
from ..order.order import BaseOrder, OrderStatus

logger = logging.getLogger(__name__)

# https://www.betfair.com/aboutUs/Betfair.Charges/#charges6
IMPLIED_COMMISSION_RATE = 0.03
PENDING_STATUS = [
    OrderStatus.PENDING,
    OrderStatus.VIOLATION,
    OrderStatus.EXPIRED,
]
ORDER_TYPE_LIMIT = OrderTypes.LIMIT
ORDER_TYPES_SP = (OrderTypes.LIMIT_ON_CLOSE, OrderTypes.MARKET_ON_CLOSE)


class Blotter:
    """
    Simple and fast class to hold all orders for
    a particular market.

    `customer_order_ref` used as the key and various
    caches available for faster access.

        blotter["abc"] = <Order>  # set
        "abc" in blotter  # contains
        orders = [o for o in blotter]  # iter
        order = blotter["abc"]  # get
    """

    def __init__(self, market_id: str):
        self.market_id = market_id
        self.active = False
        self._orders = {}  # {Order.id: Order}
        # cached lists/dicts for faster lookup
        self._trades = defaultdict(list)  # {Trade.id: [Order,]}
        self._bet_id_lookup = {}  # {Order.bet_id: Order, }
        self._live_orders = []
        self._strategy_orders = defaultdict(list)
        self._strategy_selection_orders = defaultdict(list)
        self._client_orders = defaultdict(list)
        self._client_strategy_orders = defaultdict(list)

    def get_order_bet_id(self, bet_id: str) -> Optional[BaseOrder]:
        try:
            return self._bet_id_lookup[bet_id]
        except KeyError:
            return

    def strategy_orders(
        self,
        strategy,
        order_status: Optional[List[OrderStatus]] = None,
        matched_only: Optional[bool] = None,
    ) -> list:
        """Returns all orders related to a strategy."""
        orders = self._strategy_orders[strategy]
        if order_status:
            orders = [o for o in orders if o.status in order_status]
        if matched_only:
            orders = [o for o in orders if o.size_matched > 0]
        return orders

    def strategy_selection_orders(
        self,
        strategy,
        selection_id: int,
        handicap: float = 0,
        order_status: Optional[List[OrderStatus]] = None,
        matched_only: Optional[bool] = None,
    ) -> list:
        """Returns all orders related to a strategy selection."""
        orders = self._strategy_selection_orders[(strategy, selection_id, handicap)]
        if order_status:
            orders = [o for o in orders if o.status in order_status]
        if matched_only:
            orders = [o for o in orders if o.size_matched > 0]
        return orders

    def client_orders(
        self,
        client,
        order_status: Optional[List[OrderStatus]] = None,
        matched_only: Optional[bool] = None,
    ) -> list:
        orders = self._client_orders[client]
        if order_status:
            orders = [o for o in orders if o.status in order_status]
        if matched_only:
            orders = [o for o in orders if o.size_matched > 0]
        return orders

    def client_strategy_orders(
        self,
        client,
        strategy,
        order_status: Optional[List[OrderStatus]] = None,
        matched_only: Optional[bool] = None,
    ) -> list:
        orders = self._client_strategy_orders[(client, strategy)]
        if order_status:
            orders = [o for o in orders if o.status in order_status]
        if matched_only:
            orders = [o for o in orders if o.size_matched > 0]
        return orders

    @property
    def live_orders(self) -> Iterable:
        return iter(list(self._live_orders))

    @property
    def has_live_orders(self) -> bool:
        return bool(self._live_orders)

    def process_closed_market(self, market, market_book) -> None:
        number_of_winners = len(
            [runner for runner in market_book.runners if runner.status == "WINNER"]
        )
        for order in self:
            for runner in market_book.runners:
                if (order.selection_id, order.handicap) == (
                    runner.selection_id,
                    runner.handicap,
                ):
                    order.runner_status = runner.status
                    order.market_type = market_book.market_definition.market_type
                    order.each_way_divisor = (
                        market_book.market_definition.each_way_divisor
                    )
                    if number_of_winners > market_book.number_of_winners:
                        order.number_of_dead_heat_winners = number_of_winners
                    if (
                        order.order_type.ORDER_TYPE == ORDER_TYPE_LIMIT
                        and order.order_type.price_ladder_definition == "LINE_RANGE"
                    ):
                        line_range_result = market.context.get("line_range_result")
                        if line_range_result:
                            order.line_range_result = line_range_result
                        elif order.simulated:
                            logger.warning(
                                "line_range_result unavailable, for simulation results update the market.context['line_range_result']"
                            )

    def process_cleared_orders(self, cleared_orders) -> list:
        for cleared_order in cleared_orders.orders:
            order_id = cleared_order.customer_order_ref[STRATEGY_NAME_HASH_LENGTH + 1 :]
            if order_id in self:
                self[order_id].cleared_order = cleared_order

        return [order for order in self]

    """ position """

    def market_exposure(self, strategy, market_book) -> float:
        """Returns worst-case exposure for market, which is the maximum potential loss (negative),
        arising from the worst race outcome, or the minimum potential profit (positive).
        """
        orders = self.strategy_orders(strategy)
        runners = set([order.lookup for order in orders])
        worst_possible_profits = [
            self.get_exposures(strategy, lookup) for lookup in runners
        ]
        worst_possible_profits_on_loses = [
            wpp["worst_possible_profit_on_lose"] for wpp in worst_possible_profits
        ]
        differences = [
            wpp["worst_possible_profit_on_win"] - wpp["worst_possible_profit_on_lose"]
            for wpp in worst_possible_profits
        ] + (market_book.number_of_active_runners - len(runners)) * [0]
        worst_differences = sorted(differences)[: market_book.number_of_winners]
        return sum(worst_possible_profits_on_loses) + sum(worst_differences)

    def selection_exposure(self, strategy, lookup: tuple) -> float:
        """Returns strategy/selection exposure, which is the worse-case loss arising
        from the selection either winning or losing. Can be positive or zero.
            positive = potential loss
            zero = no potential loss
        """
        exposures = self.get_exposures(strategy=strategy, lookup=lookup)
        exposure = -min(
            exposures["worst_possible_profit_on_win"],
            exposures["worst_possible_profit_on_lose"],
        )
        return max(exposure, 0.0)

    def get_exposures(self, strategy, lookup: tuple, exclusion=None) -> dict:
        """Returns strategy/selection exposures as a dict."""
        mb, ml = [], []  # matched bets, (price, size)
        ub, ul = [], []  # unmatched bets, (price, size)
        moc_win_liability = 0.0
        moc_lose_liability = 0.0
        for order in self.strategy_selection_orders(strategy, *lookup[1:]):
            if order == exclusion:
                continue
            if order.status in PENDING_STATUS:
                continue
            if order.order_type.ORDER_TYPE == ORDER_TYPE_LIMIT:
                _size_matched = order.size_matched  # cache
                _order_side = order.side
                if _size_matched:
                    if order.order_type.price_ladder_definition == "LINE_RANGE":
                        average_price_matched = 2.0
                    else:
                        average_price_matched = order.average_price_matched
                    if _order_side == "BACK":
                        mb.append((average_price_matched, _size_matched))
                    else:
                        ml.append((average_price_matched, _size_matched))
                if not order.complete:
                    _size_remaining = order.size_remaining  # cache
                    if order.order_type.price_ladder_definition == "LINE_RANGE":
                        order_type_price = 2.0
                    else:
                        order_type_price = order.order_type.price
                    if order_type_price and _size_remaining:
                        if _order_side == "BACK":
                            ub.append((order_type_price, _size_remaining))
                        else:
                            ul.append((order_type_price, _size_remaining))
            elif order.order_type.ORDER_TYPE in ORDER_TYPES_SP:
                if order.side == "BACK":
                    moc_lose_liability -= order.order_type.liability
                else:
                    moc_win_liability -= order.order_type.liability
            else:
                raise ValueError(
                    "Unexpected order type: %s" % order.order_type.ORDER_TYPE
                )
        matched_exposure = calculate_matched_exposure(mb, ml)
        unmatched_exposure = calculate_unmatched_exposure(ub, ul)

        worst_possible_profit_on_win = (
            matched_exposure[0] + unmatched_exposure[0] + moc_win_liability
        )
        worst_possible_profit_on_lose = (
            matched_exposure[1] + unmatched_exposure[1] + moc_lose_liability
        )

        return {
            "matched_profit_if_win": matched_exposure[0],
            "matched_profit_if_lose": matched_exposure[1],
            "worst_potential_unmatched_profit_if_win": unmatched_exposure[0],
            "worst_potential_unmatched_profit_if_lose": unmatched_exposure[1],
            "worst_possible_profit_on_win": worst_possible_profit_on_win,
            "worst_possible_profit_on_lose": worst_possible_profit_on_lose,
        }

    """ getters / setters """

    def complete_order(self, order) -> None:
        self._live_orders.remove(order)

    def has_order(self, customer_order_ref: str) -> bool:
        return customer_order_ref in self._orders

    def has_trade(self, trade_id: str) -> bool:
        return trade_id in self._trades

    __contains__ = has_order

    def __setitem__(self, customer_order_ref: str, order) -> None:
        self.active = True
        self._orders[customer_order_ref] = order
        self._bet_id_lookup[order.bet_id] = order
        self._live_orders.append(order)
        strategy = order.trade.strategy
        self._trades[order.trade.id].append(order)
        self._strategy_orders[strategy].append(order)
        self._strategy_selection_orders[
            (strategy, order.selection_id, order.handicap)
        ].append(order)
        client = order.client
        self._client_orders[client].append(order)
        self._client_strategy_orders[(client, strategy)].append(order)

    def __getitem__(self, customer_order_ref: str):
        return self._orders[customer_order_ref]

    def __iter__(self) -> Iterable[BaseOrder]:
        return iter(list(self._orders.values()))

    def __len__(self) -> int:
        return len(self._orders)



import unittest
from unittest import mock

from flumine.order.ordertype import (
    ExchangeType,
    OrderTypes,
    BaseOrderType,
    LimitOrder,
    LimitOnCloseOrder,
    MarketOnCloseOrder,
)


class BaseOrderTypeTest(unittest.TestCase):
    def setUp(self) -> None:
        self.order_type = BaseOrderType()

    def test_init(self):
        self.assertIsNone(self.order_type.EXCHANGE)
        self.assertIsNone(self.order_type.ORDER_TYPE)

    def test_place_instruction(self):
        with self.assertRaises(NotImplementedError):
            self.order_type.place_instruction()

    def test_info(self):
        with self.assertRaises(NotImplementedError):
            assert self.order_type.info


class LimitOrderTest(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_line_range_info = mock.Mock()
        self.order_type = LimitOrder(
            1.01,
            2,
            "PERSIST",
            "YES",
            34.5,
            "NO",
            2,
            "FINEST",
            self.mock_line_range_info,
        )

    def test_init(self):
        self.assertEqual(self.order_type.EXCHANGE, ExchangeType.BETFAIR)
        self.assertEqual(self.order_type.ORDER_TYPE, OrderTypes.LIMIT)
        self.assertEqual(self.order_type.price, 1.01)
        self.assertEqual(self.order_type.size, 2)
        self.assertEqual(self.order_type.persistence_type, "PERSIST")
        self.assertEqual(self.order_type.time_in_force, "YES")
        self.assertEqual(self.order_type.min_fill_size, 34.5)
        self.assertEqual(self.order_type.bet_target_type, "NO")
        self.assertEqual(self.order_type.bet_target_size, 2)
        self.assertEqual(self.order_type.price_ladder_definition, "FINEST")
        self.assertEqual(self.order_type.line_range_info, self.mock_line_range_info)

    def test_place_instruction(self):
        self.assertEqual(
            self.order_type.place_instruction(),
            {
                "betTargetSize": 2,
                "betTargetType": "NO",
                "minFillSize": 34.5,
                "persistenceType": "PERSIST",
                "price": 1.01,
                "size": 2,
                "timeInForce": "YES",
            },
        )

    def test_info(self):
        self.assertEqual(
            self.order_type.info,
            {
                "bet_target_size": 2,
                "bet_target_type": "NO",
                "min_fill_size": 34.5,
                "order_type": "Limit",
                "persistence_type": "PERSIST",
                "price": 1.01,
                "size": 2,
                "time_in_force": "YES",
                "price_ladder_definition": "FINEST",
            },
        )


class LimitOnCloseOrderTest(unittest.TestCase):
    def setUp(self) -> None:
        self.order_type = LimitOnCloseOrder(64, 1.01)

    def test_init(self):
        self.assertEqual(self.order_type.EXCHANGE, ExchangeType.BETFAIR)
        self.assertEqual(self.order_type.ORDER_TYPE, OrderTypes.LIMIT_ON_CLOSE)
        self.assertEqual(self.order_type.liability, 64)
        self.assertEqual(self.order_type.price, 1.01)
        self.assertEqual(self.order_type.price_ladder_definition, "CLASSIC")

    def test_place_instruction(self):
        self.assertEqual(
            self.order_type.place_instruction(), {"price": 1.01, "liability": 64}
        )

    def test_info(self):
        self.assertEqual(
            self.order_type.info,
            {
                "liability": 64,
                "order_type": "Limit on close",
                "price": 1.01,
                "price_ladder_definition": "CLASSIC",
            },
        )


class MarketOnCloseOrderTest(unittest.TestCase):
    def setUp(self) -> None:
        self.order_type = MarketOnCloseOrder(128)

    def test_init(self):
        self.assertEqual(self.order_type.EXCHANGE, ExchangeType.BETFAIR)
        self.assertEqual(self.order_type.ORDER_TYPE, OrderTypes.MARKET_ON_CLOSE)
        self.assertEqual(self.order_type.liability, 128)

    def test_place_instruction(self):
        self.assertEqual(self.order_type.place_instruction(), {"liability": 128})

    def test_info(self):
        self.assertEqual(
            self.order_type.info, {"liability": 128, "order_type": "Market on close"}
        )
